@@ -1,7 +1,7 @@
 ;; -*- mode: emacs-lisp -*-
 ;; mew-shimbun.el --- View shimbun contents with Mew
 
-;; Copyright (C) 2001 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
+;; Copyright (C) 2001, 2002, 2003 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 
 ;; Author: TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 ;;         Hideyuki SHIRAI <shirai@meadowy.org>
@@ -48,9 +48,6 @@
 ;; (define-key mew-summary-mode-map "GR" 'mew-shimbun-re-retrieve-all)
 ;; (define-key mew-summary-mode-map "Ge" 'mew-shimbun-expire)
 ;; (define-key mew-summary-mode-map "GE" 'mew-shimbun-expire-all)
-;; (when mew-shimbun-use-unseen
-;;   (define-key mew-summary-mode-map "Gu" 'mew-shimbun-unseen-check)
-;;   (define-key mew-summary-mode-map "GU" 'mew-shimbun-unseen-remove-all))
 ;;
 
 ;;; Code:
@@ -92,7 +89,7 @@
   "*The folder where SHIMBUN are contained."
   :group 'shimbun
   :group 'mew-shimbun
-  :type 'string)
+  :type '(string :size 0))
 
 (defcustom mew-shimbun-folder-groups nil
   "*Alist of `shimbun folder name (exclude `mew-shimbun-folder')'
@@ -115,19 +112,24 @@ show below example,
   :group 'shimbun
   :group 'mew-shimbun
   :type '(repeat
-	  (cons (string :tag "Folder")
-		(repeat
-		 (cons (string :tag "Server.Group")
-		       (choice :tag "Index checking range"
-			       (const all)
-			       (const last)
-			       (integer :tag "pages")))))))
+	  (cons
+	   :format "%v" :indent 2
+	   (string :format "Folder: %v\n" :size 0)
+	   (repeat
+	    :format "%{Server.Group + Index_Checking_Range%}:\n %v%i\n"
+	    :indent 3 :sample-face underline
+	    (cons :format "%v" :indent 6
+		  (string :format "Server.Group: %v\n" :size 0)
+		  (radio :format "Range: %v " :value all
+			 (const :format "%v " all)
+			 (const :format "%v " last)
+			 (integer :format "Pages: %v\n" :size 0)))))))
 
 (defcustom mew-shimbun-db-file ".mew-shimbun-db"
   "*File name of mew-shimbun database."
   :group 'shimbun
   :group 'mew-shimbun
-  :type 'file)
+  :type '(file :size 0))
 
 (defcustom mew-shimbun-expires nil
   "*Alist of `shimbun folder name' and expire days.
@@ -139,8 +141,9 @@ Show below expire,
   :group 'shimbun
   :group 'mew-shimbun
   :type '(repeat
-	  (cons (string :tag "Folder")
-		(integer :tag "Days"))))
+	  (cons :format "%v" :indent 11
+		(string :format "Folder: %v\n" :size 0)
+		(integer :format "Days: %v\n" :size 0))))
 
 (defcustom mew-shimbun-use-expire-pack nil
   "*If non-nin, exec `pack' after expire."
@@ -163,21 +166,23 @@ show below example,
 "
   :group 'shimbun
   :group 'mew-shimbun
-  :type '(choice
-	  (const :tag "same 'mew-lisp-max-length'" nil)
-	  (integer :tag "limit for all group" :value 2000)
-	  (repeat :tag "alist folder and length"
-		  (cons (choice :tag "Folder"
-				(string :tag "folder")
-				(const :tag "other" t))
-			(choice :tag "Max length of database"
-				(integer :tag "length" :value 2000))))))
+  :type '(radio
+	  (const :tag "Same as `mew-lisp-max-length'" nil)
+	  (integer :format "Limit for all groups: %v\n"
+		   :size 0 :value 2000)
+	  (repeat :indent 4 :tag "Alist of folders and lengths"
+		  (cons :format "%v" :indent 8
+			(radio :format "%v" :value t
+			       (const :format "Other " t)
+			       (string :format "Folder: %v\n" :size 0))
+			(integer :format "Maximum length of database: %v\n"
+				 :size 0 :value 2000)))))
 
 (defcustom mew-shimbun-unknown-from "foo@bar.baz"
   "*Shimbun mail address when From header is strange."
   :group 'shimbun
   :group 'mew-shimbun
-  :type 'string)
+  :type '(string :size 0))
 
 (defcustom mew-shimbun-mark-re-retrieve mew-mark-multi
   "*Shimbun re-retrieve mark."
@@ -203,12 +208,6 @@ show below example,
   :group 'mew-shimbun
   :type 'boolean)
 
-(defcustom mew-shimbun-mark-unseen mew-mark-review
-  "*Shimbun unseen mark."
-  :group 'shimbun
-  :group 'mew-shimbun
-  :type 'character)
-
 (defcustom mew-shimbun-before-retrieve-hook nil
   "*Hook run after mew-shimbun-retrieve called."
   :group 'shimbun
@@ -225,10 +224,14 @@ show below example,
 (defconst mew-shimbun-db-buffer-name " *mew-shimbun-overview*")
 (defconst mew-shimbun-article-buffer-name " *mew-shimbun-article*")
 
-(defvar mew-shimbun-unseen-regex
-  (concat mew-regex-msg "\\("
-	  (regexp-quote (string mew-shimbun-mark-unseen))
-	  "\\)"))
+(defvar mew-shimbun-unseen-regex nil)
+
+(defmacro mew-shimbun-unseen-regex ()
+  `(or mew-shimbun-unseen-regex
+       (setq mew-shimbun-unseen-regex
+	     (concat mew-regex-msg "\\("
+		     (regexp-quote (string mew-shimbun-mark-unseen))
+		     "\\)"))))
 
 (defvar mew-shimbun-folder-regex
   (mew-folder-regex (file-name-as-directory mew-shimbun-folder)))
@@ -238,6 +241,15 @@ show below example,
 (defvar mew-shimbun-input-hist nil)
 
 ;;; Macro:
+(eval-when-compile
+  (if (fboundp 'static-if)
+      nil
+    (defmacro static-if (cond then &rest else)
+      ;; Like `if', but evaluate COND at compile time.
+      (if (eval cond)
+	  then
+	`(progn ,@else)))))
+
 (defmacro mew-shimbun-db-search-id (id)
   `(assoc ,id mew-shimbun-db))
 
@@ -267,6 +279,37 @@ show below example,
 		    (format mew-shimbun-lock-format2 group server get count sum))
   (force-mode-line-update))
 
+(static-if (fboundp 'mew-summary-visit-folder)
+    (defalias 'mew-shimbun-visit-folder 'mew-summary-visit-folder)
+  (defun mew-shimbun-visit-folder (folder)
+    (mew-summary-ls
+     (mew-summary-switch-to-folder folder))))
+
+(if (featurep 'xemacs)
+    nil
+  (eval-and-compile
+    (autoload 'ad-arglist "advice"))
+  (eval-when-compile
+    (defmacro function-max-args (function)
+      ;; Return the maximum number of arguments a function may be called with.
+      ;; The function may be any form that can be passed to `funcall',
+      ;; any special form, or any macro.
+      ;; If the function takes an arbitrary number of arguments or is
+      ;; a built-in special form, nil is returned."
+      (let ((fn (make-symbol "emulating-function-max-args-function"))
+	    (arglist (make-symbol "emulating-function-max-args-arglist")))
+	`(let* ((,fn ,function)
+		(,arglist (ad-arglist (progn
+					(while (symbolp ,fn)
+					  (setq ,fn (symbol-function ,fn)))
+					,fn))))
+	   (cond ((memq '&rest ,arglist)
+		  nil)
+		 ((memq '&optional ,arglist)
+		  (1- (length ,arglist)))
+		 (t
+		  (length ,arglist))))))))
+
 ;;; Main:
 ;;;###autoload
 (defun mew-shimbun-goto-unseen-folder ()
@@ -297,7 +340,7 @@ If called with '\\[universal-argument]', goto folder to have a few new messages.
 	      (if (get-buffer fld)
 		  (with-current-buffer fld
 		    (goto-char (point-min))
-		    (when (re-search-forward mew-shimbun-unseen-regex nil t)
+		    (when (re-search-forward (mew-shimbun-unseen-regex) nil t)
 		      (setq sbflds (cons fld sbflds))))
 		(setq cfile (mew-expand-folder fld mew-summary-cache-file))
 		(when (file-readable-p cfile)
@@ -306,7 +349,7 @@ If called with '\\[universal-argument]', goto folder to have a few new messages.
 		     mew-cs-text-for-read mew-cs-dummy
 		     (insert-file-contents cfile nil)
 		     (goto-char (point-min))
-		     (when (re-search-forward mew-shimbun-unseen-regex nil t)
+		     (when (re-search-forward (mew-shimbun-unseen-regex) nil t)
 		       (setq sbflds (cons fld sbflds))))))))))))
     (mapcar (lambda (x)
 	      (setq alst (cons (list x) alst)))
@@ -324,7 +367,7 @@ If called with '\\[universal-argument]', goto folder to have a few new messages.
       (setcar mew-shimbun-input-hist fld))
     (setq mew-input-folder-hist (cons fld mew-input-folder-hist))
     (let ((newfld (mew-summary-switch-to-folder fld)))
-      (if (eq 1 (w3m-function-max-args 'mew-summary-ls))
+      (if (eq 1 (function-max-args 'mew-summary-ls))
 	  (mew-summary-ls newfld)
 	(dont-compile;; To avoid a byte-compile warnning.
 	  (mew-summary-ls newfld newfld))))))
@@ -384,23 +427,23 @@ If called with '\\[universal-argument]', goto folder to have a few new messages.
 	 (setq dir (mew-expand-folder fld))
 	 (unless (file-directory-p dir)
 	   (mew-make-directory dir))
-	 (mew-summary-visit-folder fld)
+	 (mew-shimbun-visit-folder fld)
 	 (sit-for 0.5)
 	 (mew-rendezvous mew-summary-buffer-process)
 	 (mew-shimbun-retrieve)
 	 (unless (eq (get-buffer cfld) (current-buffer))
 	   (mew-kill-buffer (current-buffer)))))
-     (mew-summary-visit-folder cfld)
+     (mew-shimbun-visit-folder cfld)
      (message "Getting done"))))
 
 (defun mew-shimbun-retrieve-article (mua server group range fld)
   "Retrieve articles via SHIMBUN."
   (luna-define-method shimbun-mua-search-id ((mua shimbun-mew-mua) id)
-    (let ((shimbun (shimbun-mua-shimbun-internal mua)))
+    (let ((shimbun (shimbun-mua-shimbun mua)))
       (mew-shimbun-db-search-id
        (format mew-shimbun-id-format
-	       (shimbun-server-internal shimbun)
-	       (shimbun-current-group-internal shimbun)
+	       (shimbun-server shimbun)
+	       (shimbun-current-group shimbun)
 	       id))))
   (let ((shimbun (shimbun-open server mua))
 	(count 0)
@@ -448,7 +491,7 @@ If called with '\\[universal-argument]', goto folder to have a few new messages.
 	      (setq dispcount (1+ dispcount))
 	      (mew-shimbun-mode-display group server count dispcount sum))))
       (mew-summary-unlock)
-      (w3m-static-if (fboundp 'mew-folder-insert)
+      (static-if (fboundp 'mew-folder-insert)
 	  (mew-folder-insert fld)
 	(mew-local-folder-insert fld))
       (shimbun-close-group shimbun)
@@ -559,11 +602,11 @@ If called with '\\[universal-argument]', re-retrieve messages in the region."
 (defun mew-shimbun-re-retrieve-article (mua server group range fld id-msgs)
   "Re-retrieve articles via SHIMBUN."
   (luna-define-method shimbun-mua-search-id ((mua shimbun-mew-mua) id)
-    (let ((shimbun (shimbun-mua-shimbun-internal mua)))
+    (let ((shimbun (shimbun-mua-shimbun mua)))
       (mew-shimbun-db-search-id2
        (format mew-shimbun-id-format
-	       (shimbun-server-internal shimbun)
-	       (shimbun-current-group-internal shimbun)
+	       (shimbun-server shimbun)
+	       (shimbun-current-group shimbun)
 	       id))))
   (let ((shimbun (shimbun-open server mua))
 	(newcount 0) (rplcount 0) (same 0) (dispcount 0))
@@ -636,13 +679,13 @@ If called with '\\[universal-argument]', re-retrieve messages in the region."
       (when (and (file-directory-p (mew-expand-folder fld))
 		 (file-exists-p (expand-file-name mew-shimbun-db-file
 						  (mew-expand-folder fld))))
-	(mew-summary-visit-folder fld)
+	(mew-shimbun-visit-folder fld)
 	(sit-for 0.5)
 	(mew-rendezvous mew-summary-buffer-process)
 	(mew-shimbun-expire)
 	(unless (eq (get-buffer cfld) (current-buffer))
 	  (mew-kill-buffer (current-buffer)))))
-    (mew-summary-visit-folder cfld)))
+    (mew-shimbun-visit-folder cfld)))
 
 ;;;###autoload
 (defun mew-shimbun-expire ()
@@ -765,7 +808,9 @@ If called with '\\[universal-argument]', re-retrieve messages in the region."
 (defun mew-shimbun-scan-message (fld msg)
   (set-buffer-multibyte t)
   (let ((width (1- (mew-scan-width)))
-	(vec (mew-pop-scan-header)))
+	(vec (static-if (fboundp 'mew-pop-scan-header)
+		 (mew-pop-scan-header)
+	       (mew-scan-header))))
     (mew-scan-set-folder vec fld)
     (mew-scan-set-message vec msg)
     (set-buffer-multibyte nil)
@@ -956,7 +1001,7 @@ If called with '\\[universal-argument]', re-retrieve messages in the region."
 	(when (and fld msg (null part))
 	  (save-excursion
 	    (beginning-of-line)
-	    (when (looking-at mew-shimbun-unseen-regex)
+	    (when (looking-at (mew-shimbun-unseen-regex))
 	      ;; in normal or thread folder
 	      (mew-mark-unmark)
 	      (set-buffer-modified-p nil)
@@ -970,7 +1015,7 @@ If called with '\\[universal-argument]', re-retrieve messages in the region."
   (when mew-shimbun-use-unseen
     (defadvice mew-summary-cursor-postscript (before shimbun-unseen activate)
       (mew-shimbun-unseen-remove-advice))
-    
+
     (when mew-shimbun-use-unseen-cache-save
       ;; "C-cC-q"
       (defadvice mew-kill-buffer (before shimbun-cache-save activate)
@@ -981,7 +1026,7 @@ If called with '\\[universal-argument]', re-retrieve messages in the region."
 	      (set-buffer buf)
 	      (unless (mew-summary-folder-dir-newp)
 		(mew-summary-folder-cache-save))))))
-      
+
       ;; "Q" or exit Emacs
       (defadvice mew-mark-clean-up (before shimbun-cache-save activate)
 	(save-excursion
