@@ -1,6 +1,6 @@
 ;;; mime-w3m.el --- mime-view content filter for text
 
-;; Copyright (C) 2001, 2002 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
+;; Copyright (C) 2001, 2002, 2003 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 
 ;; Author: TSUCHIYA Masatoshi <tsuchiya@namazu.org>,
 ;;         Akihiro Arisawa    <ari@mbf.sphere.ne.jp>
@@ -54,7 +54,17 @@ When this option is equal to `default',
 to decide whether inline images are displayed."
   :group 'w3m
   :group 'mime-view
-  :type 'boolean)
+  :type '(radio (const :format "%v " nil)
+		(sexp :format "non-nil "
+		      :match
+		      (lambda (widget value)
+			(and value (not (eq value 'default))))
+		      :value-to-internal
+		      (lambda (widget value)
+			(if (and value (not (equal value "default")))
+			    (widget-sexp-value-to-internal widget value)
+			  "t")))
+		(const default)))
 
 (defcustom mime-w3m-safe-url-regexp "\\`cid:"
   "*Regexp that matches safe url names.
@@ -65,7 +75,7 @@ this to the regexp which matches the safe url names.  The value of the
 variable `w3m-safe-url-regexp' will be bound with this value.  You may
 set this value to nil if you consider all the urls to be safe."
   :group 'mime-w3m
-  :type '(choice (regexp :tag "Regexp")
+  :type '(choice (regexp :format "%t: %v\n" :size 0)
 		 (const :tag "All URLs are safe" nil)))
 
 (defcustom mime-w3m-setup-hook nil
@@ -124,15 +134,17 @@ set this value to nil if you consider all the urls to be safe."
 	   (font-set-face-background 'default color (current-buffer))))
     (cons 'progn body)))
 
-(eval-and-compile
-  (unless (or (featurep 'xemacs)
-	      (>= emacs-major-version 21))
-    (defvar mime-w3m-mode-map nil
-      "Keymap for text/html part rendered by `mime-w3m-preview-text/html'.
+(unless (or (featurep 'xemacs)
+	    (>= emacs-major-version 21))
+  (defvar mime-w3m-mode-map nil
+    "Keymap for text/html part rendered by `mime-w3m-preview-text/html'.
 This map is overwritten by `mime-w3m-local-map-property' based on the
 value of `w3m-minor-mode-map'.  Therefore, in order to add some
 commands to this map, add them to `w3m-minor-mode-map' instead of this
-map.")))
+map."))
+
+(eval-when-compile
+  (defvar mime-w3m-mode-map))
 
 (defsubst mime-w3m-local-map-property ()
   (if (or (featurep 'xemacs)
@@ -169,11 +181,14 @@ map.")))
 	     (w3m-region p (point-max)
 			 (and (stringp xref)
 			      (string-match "\\`http://" xref)
-			      xref))
+			      xref)
+			 (mime-content-type-parameter
+			  (mime-entity-content-type entity)
+			  "charset"))
 	     (add-text-properties p (point-max)
 				  (nconc (mime-w3m-local-map-property)
 					 '(text-rendered-by-mime-w3m t))))
-	 (error (message (format "%s" err))))))))
+	 (error (message "%s" err)))))))
 
 ;; To avoid byte-compile warning in `mime-w3m-cid-retrieve'.
 (autoload 'mime-uri-parse-cid "mime-parse")
@@ -189,15 +204,14 @@ map.")))
       (mime-entity-type/subtype entity))))
 
 (let (current-load-list)
-  (defadvice kill-new (before strip-keymap-properties-from-kill
-			      (string &optional replace) activate)
+  (defadvice kill-new (before strip-keymap-properties-from-kill activate)
     "Advised by emacs-w3m.
 Strip `keymap' or `local-map' properties from a killed string."
-    (if (text-property-any 0 (length string)
-			   'text-rendered-by-mime-w3m t string)
-	(remove-text-properties 0 (length string)
+    (if (text-property-any 0 (length (ad-get-arg 0))
+			   'text-rendered-by-mime-w3m t (ad-get-arg 0))
+	(remove-text-properties 0 (length (ad-get-arg 0))
 				'(keymap nil local-map nil)
-				string))))
+				(ad-get-arg 0)))))
 
 (mime-w3m-insinuate)
 
