@@ -1,9 +1,10 @@
 ;;; w3m-e21.el --- The stuffs to use emacs-w3m on Emacs-21
 
-;; Copyright (C) 2001, 2002 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
+;; Copyright (C) 2001, 2002, 2003 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 
 ;; Authors: Yuuichi Teranishi  <teranisi@gohome.org>,
-;;          TSUCHIYA Masatoshi <tsuchiya@namazu.org>
+;;          TSUCHIYA Masatoshi <tsuchiya@namazu.org>,
+;;          Katsumi Yamaoka    <yamaoka@jpl.org>
 ;; Keywords: w3m, WWW, hypermedia
 
 ;; This file is a part of emacs-w3m.
@@ -40,34 +41,41 @@
 (require 'w3m-util)
 (require 'w3m-proc)
 (require 'w3m-image)
+(require 'w3m-favicon)
 (require 'w3m-fsf)
+(require 'w3m-ccl)
 (require 'wid-edit)
 
 ;; Functions and variables which should be defined in the other module
 ;; at run-time.
 (eval-when-compile
+  (defvar w3m-current-process)
   (defvar w3m-current-url)
   (defvar w3m-current-title)
-  (defvar w3m-current-process)
-  (defvar w3m-current-buffer)
   (defvar w3m-display-inline-images)
+  (defvar w3m-favicon-image)
+  (defvar w3m-form-use-fancy-faces)
   (defvar w3m-icon-directory)
   (defvar w3m-mode-map)
-  (defvar w3m-profile-directory)
+  (defvar w3m-modeline-process-status-on)
+  (defvar w3m-process-queue)
+  (defvar w3m-show-graphic-icons-in-header-line)
+  (defvar w3m-show-graphic-icons-in-mode-line)
   (defvar w3m-toolbar)
   (defvar w3m-toolbar-buttons)
-  (defvar w3m-use-tab)
+  (defvar w3m-use-favicon)
   (defvar w3m-use-header-line)
+  (defvar w3m-use-tab)
   (defvar w3m-work-buffer-name)
-  (defvar w3m-history)
-  (defvar w3m-history-flat)
-  (defvar w3m-form-use-fancy-faces)
-  (defvar w3m-async-exec)
-  (autoload 'w3m-expand-url "w3m")
-  (autoload 'w3m-retrieve "w3m")
+  (defvar w3m-work-buffer-list)
   (autoload 'w3m-image-type "w3m")
-  (autoload 'w3m-load-list "w3m")
-  (autoload 'w3m-save-list "w3m"))
+  (autoload 'w3m-retrieve "w3m"))
+
+(eval-and-compile
+  (unless (fboundp 'frame-current-scroll-bars)
+    (defalias 'frame-current-scroll-bars 'ignore))
+  (unless (fboundp 'window-fringes)
+    (defalias 'window-fringes 'ignore)))
 
 ;;; Coding system.
 
@@ -78,6 +86,72 @@ CODING-SYSTEM, DECODER and ENCODER must be symbol."
   (make-coding-system coding-system 4 mnemonic docstring
 		      (cons decoder encoder)))
 
+(eval-and-compile
+  (defconst w3m-ccl-get-ucs-codepoint-with-emacs-unicode
+    `((if (r1 == ,(charset-id 'latin-iso8859-1))
+	  ((r1 = (r0 + 128)))
+	(if (r1 == ,(charset-id 'mule-unicode-0100-24ff))
+	    ((r1 = ((((r0 & #x3f80) >> 7) - 32) * 96))
+	     (r0 &= #x7f)
+	     (r1 += (r0 + 224)))	; 224 == -32 + #x0100
+	  (if (r1 == ,(charset-id 'mule-unicode-2500-33ff))
+	      ((r1 = ((((r0 & #x3f80) >> 7) - 32) * 96))
+	       (r0 &= #x7f)
+	       (r1 += (r0 + 9440)))	; 9440 == -32 + #x2500
+	    (if (r1 == ,(charset-id 'mule-unicode-e000-ffff))
+		((r1 = ((((r0 & #x3f80) >> 7) - 32) * 96))
+		 (r0 &= #x7f)
+		 (r1 += (r0 + 57312)))	; 57312 == -32 + #xe000
+	      ,(if (fboundp 'ccl-compile-lookup-character)
+		   '((lookup-character utf-subst-table-for-encode r1 r0)
+		     (if (r7 == 0)	; lookup failed
+			 (r1 = #xfffd)))
+		 '((r1 = #xfffd)))))))
+      (if (r1 == #xfffd)
+	  (write-repeat ?~)		; unknown character.
+	(r0 = r1)))
+    "CCL program to convert multibyte char to ucs with emacs-unicode."))
+
+(if (get 'utf-translation-table-for-encode 'translation-table-id)
+    ;; For Emacs 21.3 and later.
+    (eval
+     ;; Compute the ccl program in each time to load this module in order to
+     ;; keep the compatibility of this module between Emacs 21.3 and later.
+     '(define-ccl-program w3m-euc-japan-encoder
+	`(4
+	  (loop
+	   ,@w3m-ccl-write-euc-japan-character
+	   (translate-character utf-translation-table-for-encode r1 r0)
+	   ,@w3m-ccl-get-ucs-codepoint-with-emacs-unicode
+	   ,@w3m-ccl-generate-ncr))))
+  ;; For Emacs 21.[12] that does not have `utf-translation-table-for-encode'.
+  (define-ccl-program w3m-euc-japan-encoder
+    `(4
+      (loop
+       ,@w3m-ccl-write-euc-japan-character
+       ,@w3m-ccl-get-ucs-codepoint-with-emacs-unicode
+       ,@w3m-ccl-generate-ncr))))
+
+(if (get 'utf-translation-table-for-encode 'translation-table-id)
+    ;; For Emacs 21.3 and later.
+    (eval
+     ;; Compute the ccl program in each time to load this module in order to
+     ;; keep the compatibility of this module between Emacs 21.3 and later.
+     '(define-ccl-program w3m-iso-latin-1-encoder
+	`(4
+	  (loop
+	   ,@w3m-ccl-write-iso-latin-1-character
+	   (translate-character utf-translation-table-for-encode r1 r0)
+	   ,@w3m-ccl-get-ucs-codepoint-with-emacs-unicode
+	   ,@w3m-ccl-generate-ncr))))
+  ;; For Emacs 21.[12] that does not have `utf-translation-table-for-encode'.
+  (define-ccl-program w3m-iso-latin-1-encoder
+    `(4
+      (loop
+       ,@w3m-ccl-write-iso-latin-1-character
+       ,@w3m-ccl-get-ucs-codepoint-with-emacs-unicode
+       ,@w3m-ccl-generate-ncr))))
+
 (unless (fboundp 'w3m-ucs-to-char)
   (defun w3m-ucs-to-char (codepoint)
     (or (decode-char 'ucs codepoint) ?~)))
@@ -286,8 +360,9 @@ Buffer string between BEG and END are replaced with IMAGE."
 (defvar w3m-e21-toolbar-configurations
   '((auto-resize-tool-bars       . t)
     (auto-raise-tool-bar-buttons . t)
-    (tool-bar-button-margin      . 0)
-    (tool-bar-button-relief      . 2)))
+    ;;(tool-bar-button-margin      . 0)
+    ;;(tool-bar-button-relief      . 2)
+    ))
 
 (defun w3m-e21-setup-toolbar (keymap defs)
   (let ((configs w3m-e21-toolbar-configurations)
@@ -342,222 +417,41 @@ Buffer string between BEG and END are replaced with IMAGE."
 
 (defalias 'w3m-update-toolbar 'ignore)
 
-;;; favicon
-(defcustom w3m-favicon-size nil
-  "*Size of favicon. This value is used as geometry argument for `convert'."
-  :group 'w3m
-  :type 'string)
-
-(defconst w3m-favicon-name "favicon.ico"
-  "The favicon name.")
-
-(defvar w3m-current-favicon-data nil
-  "A cons cell of (IMAGE-DATA . IMAGE-TYPE).")
-(defvar w3m-current-favicon-image nil)
-(defvar w3m-favicon-converted nil)
-(make-variable-buffer-local 'w3m-current-favicon-data)
-(make-variable-buffer-local 'w3m-current-favicon-image)
-(make-variable-buffer-local 'w3m-favicon-converted)
-(add-hook 'w3m-display-functions 'w3m-setup-favicon)
-
-(defun w3m-favicon-usable-p ()
-  "Check whether ImageMagick's `convert' supports a Windoze ico format in
-a large number of bits per pixel."
-  (let ((xpm (condition-case nil
-		 (w3m-imagick-convert-data
-		  (string 0 0 1 0 1 0 2 1 0 0 1 0 24 0 52 0
-			  0 0 22 0 0 0 40 0 0 0 2 0 0 0 2 0
-			  0 0 1 0 24 0 0 0 0 0 0 0 0 0 0 0
-			  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
-			  0 255 255 255 0 0 0 0 0 0)
-		  "ico" "xpm")
-	       (error nil))))
-    (and xpm (string-match "\"2 1 2 1\"" xpm) t)))
-
-(defcustom w3m-use-favicon t
-  "*If non-nil, use favicon.
-It will be set to nil automatically if ImageMagick's
-`convert' does not support a ico format.  You can inhibit the
-use of ImageMagick absolutely by setting this option to nil."
-  :get (lambda (symbol)
-	 (and (not noninteractive)
-	      (default-value symbol)
-	      (w3m-favicon-usable-p)))
-  :set (lambda (symbol value)
-	 (custom-set-default symbol
-			     (and (not noninteractive)
-				  value
-				  (w3m-favicon-usable-p))))
-  :group 'w3m
-  :type 'boolean)
-
-(defcustom w3m-favicon-use-cache-file nil
-  "*If non-nil, use favicon cache file."
-  :group 'w3m
-  :type 'boolean)
-
-(defcustom w3m-favicon-cache-file nil
-  "Filename of saving favicon cache."
-  :group 'w3m
-  :type 'file)
-
-(defcustom w3m-favicon-cache-expire-wait (* 30 24 60 60)
-  "*The cache will be expired after specified seconds passed since retrieval.
-If this variable is nil, never expired."
-  :group 'w3m
-  :type 'integer)
-
-(defcustom w3m-favicon-type 'xpm
-  "*Image type of display favicon."
-  :group 'w3m
-  :type (cons 'choice
-	      (mapcar (lambda (x)
-			`(const :tag ,(symbol-name x) ,x))
-		      (delq 'postscript
-			    (delq 'xbm
-				  (delq 'pbm (copy-sequence image-types)))))))
-
-(defvar w3m-favicon-cache-data nil
-  "A list of favicon cache (internal variable).
-Each information is a list whose elements are:
- 0. URL
- 1. Favicon
- 2. Retrieved date")
-
-(defmacro w3m-favicon-cache-p (url)
-  `(assoc ,url w3m-favicon-cache-data))
-
-(defmacro w3m-favicon-cache-favicon (url)
-  `(let ((data (nth 1 (assoc ,url w3m-favicon-cache-data))))
-     (if (stringp data) (cons data 'ico) data)))
-
-(defmacro w3m-favicon-cache-retrieved (url)
-  `(nth 2 (assoc ,url w3m-favicon-cache-data)))
-
-(defun w3m-setup-favicon (url)
-  (setq w3m-current-favicon-data nil
-	w3m-current-favicon-image nil
-	w3m-favicon-converted nil)
-  (when (and w3m-use-favicon
-	     w3m-current-url
-	     (w3m-image-type-available-p 'xpm)
-	     (w3m-image-type-available-p w3m-favicon-type))
-    (cond
-     ((string-match "\\`about://\\([^/]+\\)/" url)
-      (let ((icon (intern-soft (concat "w3m-about-" (match-string 1 url)
-				       "-favicon"))))
-	(if (and (fboundp 'base64-decode-string)
-		 icon)
-	    (with-current-buffer w3m-current-buffer
-	      (setq w3m-current-favicon-data
-		    (cons (eval (list 'base64-decode-string
-				      (symbol-value icon)))
-			  'ico))))))
-     ((string-match "\\`https?://" url)
-      (w3m-retrieve-favicon
-       (or (symbol-value 'w3m-icon-data)
-	   (cons
-	    (w3m-expand-url (concat "/" w3m-favicon-name)
-			    url)
-	    'ico))
-       w3m-current-buffer)))))
-
-(defun w3m-buffer-favicon (buffer)
-  (with-current-buffer buffer
-    (when (and w3m-current-favicon-data (car w3m-current-favicon-data))
-      (or w3m-current-favicon-image
-	  w3m-favicon-converted
-	  (lexical-let ((height (number-to-string (frame-char-height)))
-			(buffer buffer))
-	    (w3m-process-with-null-handler
-	      (setq w3m-favicon-converted t)
-	      (w3m-process-do
-		  (img (w3m-imagick-start-convert-data
-			handler
-			(car w3m-current-favicon-data)
-			(symbol-name (cdr w3m-current-favicon-data))
-			(symbol-name w3m-favicon-type)
-			"-geometry" (or w3m-favicon-size
-					(concat height "x" height))))
-		(with-current-buffer buffer
-		  (if img
-		      (setq w3m-current-favicon-image
-			    (create-image img
-					  w3m-favicon-type
-					  t
-					  :ascent 'center))
-		    (setq w3m-current-favicon-data nil))))))))))
-
-(defun w3m-retrieve-favicon (pair target &optional handler)
-  (if (and (w3m-favicon-cache-p (car pair))
-	   (or (null w3m-favicon-cache-expire-wait)
-	       (< (- (float-time)
-		     (float-time (w3m-favicon-cache-retrieved
-				  (car pair))))
-		  w3m-favicon-cache-expire-wait)))
-      (setq w3m-current-favicon-data
-	    (w3m-favicon-cache-favicon (car pair)))
-    (lexical-let ((pair pair)
-		  (target target))
-      (w3m-process-do-with-temp-buffer
-	  (ok (w3m-retrieve (car pair) 'raw nil nil nil handler))
-	(let (idata)
-	  (when ok
-	    (setq idata (buffer-string))
-	    (with-current-buffer target
-	      (setq w3m-current-favicon-data (cons idata (cdr pair)))))
-	  (push (list (car pair) (and idata (cons idata (cdr pair)))
-		      (current-time))
-		w3m-favicon-cache-data))))))
-
-(defun w3m-favicon-save-cache-file ()
-  (when w3m-favicon-use-cache-file
-    (w3m-save-list (or w3m-favicon-cache-file
-		       (expand-file-name ".favicon" w3m-profile-directory))
-		   w3m-favicon-cache-data 'binary)))
-
-(defun w3m-favicon-load-cache-file ()
-  (when (and w3m-favicon-use-cache-file
-	     (null w3m-favicon-cache-data))
-    (setq w3m-favicon-cache-data
-	  (w3m-load-list
-	   (or w3m-favicon-cache-file
-	       (expand-file-name ".favicon" w3m-profile-directory))
-	   'binary))))
-
-(add-hook 'w3m-arrived-setup-functions 'w3m-favicon-load-cache-file)
-(add-hook 'w3m-arrived-shutdown-functions 'w3m-favicon-save-cache-file)
-
 ;;; Header line & Tabs
 (defcustom w3m-tab-width 16
   "w3m tab width."
   :group 'w3m
-  :type 'integer)
+  :set (lambda (symbol value)
+	 (custom-set-default symbol
+			     (if (and (numberp value) (> value 0))
+				 value
+			       16)))
+  :type '(integer :size 0))
 
 (defface w3m-tab-unselected-face
   '((((type x w32 mac) (class color))
      :background "Gray50" :foreground "Gray20"
-     :underline "Gray85" :box (:line-width -1 :style released-button))
+     :box (:line-width -1 :style released-button))
     (((class color))
-     (:background "cyan" :foreground "black" :underline "blue")))
+     (:background "blue" :foreground "black")))
   "*Face to fontify unselected tabs."
   :group 'w3m-face)
 
 (defface w3m-tab-unselected-retrieving-face
   '((((type x w32 mac) (class color))
      :background "Gray50" :foreground "OrangeRed"
-     :underline "Gray85" :box (:line-width -1 :style released-button))
+     :box (:line-width -1 :style released-button))
     (((class color))
-     (:background "cyan" :foreground "OrangeRed" :underline "blue")))
+     (:background "blue" :foreground "OrangeRed")))
   "*Face to fontify unselected tabs which are retrieving their pages."
   :group 'w3m-face)
 
 (defface w3m-tab-selected-face
   '((((type x w32 mac) (class color))
      :background "Gray85" :foreground "black"
-     :underline "Gray85" :box (:line-width -1 :style released-button))
+     :box (:line-width -1 :style released-button))
     (((class color))
-     (:background "blue" :foreground "black" :underline "blue"))
+     (:background "cyan" :foreground "black"))
     (t (:underline t)))
   "*Face to fontify selected tab."
   :group 'w3m-face)
@@ -565,43 +459,47 @@ Each information is a list whose elements are:
 (defface w3m-tab-selected-retrieving-face
   '((((type x w32 mac) (class color))
      :background "Gray85" :foreground "red"
-     :underline "Gray85" :box (:line-width -1 :style released-button))
+     :box (:line-width -1 :style released-button))
     (((class color))
-     (:background "blue" :foreground "red" :underline "blue"))
+     (:background "cyan" :foreground "red"))
     (t (:underline t)))
   "*Face to fontify selected tab which is retrieving its page."
   :group 'w3m-face)
 
 (defface w3m-tab-background-face
   '((((type x w32 mac) (class color))
-     :background "LightSteelBlue" :foreground "black"
-     :underline "Gray85")
+     :background "LightSteelBlue" :foreground "black")
     (((class color))
-     (:background "white" :foreground "black" :underline "blue")))
+     (:background "white" :foreground "black")))
   "*Face to fontify background of tab line."
   :group 'w3m-face)
 
+(defvar w3m-modeline-spinner-map nil
+  "Keymap used on the spinner in the mode-line.")
+
+(defvar w3m-spinner-map-help-echo "mouse-2 kills the current process"
+  "String used for the :help-echo property on the spinner.")
+
 (defun w3m-setup-header-line ()
   (setq header-line-format
-	(cond
-	 (w3m-use-tab
-	  (list '(:eval (w3m-tab-line))))
-	 (w3m-use-header-line
-	  (list
-	   (propertize
-	    "Location: "
-	    'face 'w3m-header-line-location-title-face)
-	   '(:eval
-	     (propertize
-	      (if (stringp w3m-current-url)
-		  (replace-regexp-in-string "%" "%%" w3m-current-url)
-		"")
-	      'face 'w3m-header-line-location-content-face
-	      'local-map (let ((map (make-sparse-keymap)))
-			   (define-key map [header-line mouse-2]
-			     'w3m-goto-url)
-			   map)
-	      'help-echo "mouse-2 prompts to input URL")))))))
+	(cond (w3m-use-tab
+	       '(:eval (w3m-tab-line)))
+	      (w3m-use-header-line
+	       (list
+		(propertize
+		 "Location: "
+		 'face 'w3m-header-line-location-title-face)
+		'(:eval
+		  (propertize
+		   (if (stringp w3m-current-url)
+		       (replace-regexp-in-string "%" "%%" w3m-current-url)
+		     "")
+		   'face 'w3m-header-line-location-content-face
+		   'local-map (let ((map (make-sparse-keymap)))
+				(define-key map [header-line mouse-2]
+				  'w3m-goto-url)
+				map)
+		   'help-echo "mouse-2 prompts to input URL")))))))
 
 (defun w3m-tab-drag-mouse-function (event buffer)
   (let ((window (posn-window (event-end event)))
@@ -616,81 +514,294 @@ Each information is a list whose elements are:
 				   (cdr (cdr mpos))
 				   (car mpos)))))
     (when window
-      (unless (string= (buffer-name (window-buffer window))
-		       buffer)
+      (unless (eq (window-buffer window) buffer)
 	(select-window window)
 	(switch-to-buffer buffer)))))
 
-(defun w3m-tab-make-keymap (buffer)
-  (let ((map (make-sparse-keymap))
-	(drag-action `(lambda (e) (interactive "e")
-			(w3m-tab-drag-mouse-function
-			 e ,(buffer-name buffer))))
-	(up-action `(lambda (e) (interactive "e")
-		      (switch-to-buffer ,(buffer-name buffer)))))
-    (define-key map [header-line down-mouse-1] 'ignore)
-    (define-key map [header-line down-mouse-2] 'ignore)
-    (define-key map [header-line drag-mouse-1] drag-action)
-    (define-key map [header-line drag-mouse-2] drag-action)
-    (define-key map [header-line mouse-1] up-action)
-    (define-key map [header-line mouse-2] up-action)
-    map))
+(defvar w3m-tab-map nil)
+(make-variable-buffer-local 'w3m-tab-map)
+
+(defvar w3m-tab-spinner-map nil)
+(make-variable-buffer-local 'w3m-tab-spinner-map)
+
+(defun w3m-tab-make-keymap ()
+  (unless w3m-tab-map
+    (setq w3m-tab-map (make-sparse-keymap))
+    (let* ((buffer (current-buffer))
+	   (drag-action `(lambda (e)
+			   (interactive "e")
+			   (w3m-tab-drag-mouse-function e ,buffer)))
+	   (up-action `(lambda (e)
+			 (interactive "e")
+			 (switch-to-buffer ,buffer))))
+      (define-key w3m-tab-map [header-line down-mouse-1] 'ignore)
+      (define-key w3m-tab-map [header-line down-mouse-2] 'ignore)
+      (define-key w3m-tab-map [header-line drag-mouse-1] drag-action)
+      (define-key w3m-tab-map [header-line drag-mouse-2] drag-action)
+      (define-key w3m-tab-map [header-line mouse-1] up-action)
+      (define-key w3m-tab-map [header-line mouse-2] up-action)))
+  (unless w3m-tab-spinner-map
+    (setq w3m-tab-spinner-map (make-sparse-keymap))
+    (define-key w3m-tab-spinner-map [header-line mouse-2]
+      `(lambda (e)
+	 (interactive "e")
+	 (save-current-buffer
+	   ;; Why the `(w3m-process-stop BUFFER)' doesn't work?
+	   (set-buffer ,(current-buffer))
+	   (call-interactively 'w3m-process-stop))))))
+
+(defvar w3m-tab-line-format nil
+  "Internal variable used to keep contents to be shown in the header-line.")
+
+(defvar w3m-tab-timer nil
+  "Internal variable used to say time has not gone by after the tab-line
+was updated last time.  It is used to control the `w3m-tab-line'
+function running too frequently, set by the function itself and
+cleared by a timer.")
+
+(defvar w3m-tab-half-space
+  (propertize " " 'display '(space :width 0.5))
+  "The space of half width.")
+
+(defvar w3m-tab-separator
+  (propertize " "
+	      'face 'w3m-tab-background-face
+	      'display '(space :width 0.5))
+  "String used to separate tabs.")
 
 (defun w3m-tab-line ()
-  (let* ((current (current-buffer))
-	 (buffers (w3m-list-buffers))
-	 (width (if (> (* (length buffers) (+ 5 w3m-tab-width))
-		       (window-width))
-		    (max (- (/ (window-width) (length buffers)) 5) 1)
-		  w3m-tab-width)))
-    (concat
-     (save-current-buffer
-       (mapconcat
-	(lambda (buffer)
-	  (let ((title (w3m-buffer-title buffer))
-		(favicon (if w3m-use-favicon (w3m-buffer-favicon buffer))))
+  (or (and w3m-tab-timer w3m-tab-line-format)
+      (let* ((current (current-buffer))
+	     (buffers (w3m-list-buffers))
+	     (breadth 1)
+	     (fringes (window-fringes))
+	     (width (+ (window-width)
+		       (/ (float (+ (or (car fringes) 0)
+				    (or (nth 1 fringes) 0)))
+			  (frame-char-width))
+		       ;; Assume that the vertical scroll-bar has
+		       ;; the width of two space characters.
+		       (if (car (frame-current-scroll-bars)) 2 0)))
+	     (nbuf (length buffers))
+	     (graphic (and window-system
+			   w3m-show-graphic-icons-in-header-line))
+	     (margin (if window-system
+			 (+ (if graphic 3.0 0.5)
+			    ;; Right and left shadows.
+			    (/ 2.0 (frame-char-width)))
+		       1))
+	     (spinner (when w3m-process-queue
+			(w3m-make-spinner-image)))
+	     buffer title data datum process favicon keymap face icon line)
+	(setq w3m-tab-timer
+	      (run-at-time 0.1 nil
+			   (lambda (window)
+			     (setq w3m-tab-timer nil)
+			     (if (and (eq (selected-window) window)
+				      w3m-process-queue)
+				 ;; Wobble the window size to force
+				 ;; redisplay of the header-line.
+				 (let ((window-min-height 0))
+				   (shrink-window 1)
+				   (enlarge-window 1))))
+			   (get-buffer-window current t)))
+	(save-current-buffer
+	  (while buffers
+	    (set-buffer (setq buffer (pop buffers)))
+	    (setq breadth
+		  (max breadth
+		       ;; There may be a wide character in the beginning of
+		       ;; the title.
+		       (char-width (aref (setq title (w3m-current-title)) 0))))
+	    (push (list (eq current buffer)
+			w3m-current-process
+			title
+			(when w3m-use-favicon w3m-favicon-image)
+			w3m-tab-map)
+		  data)))
+	(setq width (if (> (* nbuf (+ margin w3m-tab-width)) width)
+			(max (truncate (- (/ width nbuf) margin)) breadth)
+		      w3m-tab-width))
+	(while data
+	  (setq datum (pop data)
+		current (car datum)
+		process (nth 1 datum)
+		title (nth 2 datum)
+		favicon (nth 3 datum)
+		keymap (nth 4 datum)
+		face (if process
+			 (if current
+			     'w3m-tab-selected-retrieving-face
+			   'w3m-tab-unselected-retrieving-face)
+		       (if current
+			   'w3m-tab-selected-face
+			 'w3m-tab-unselected-face))
+		icon (when graphic
+		       (cond
+			(process
+			 (when spinner
+			   (propertize
+			    " "
+			    'display spinner
+			    'face face
+			    'local-map w3m-tab-spinner-map
+			    'help-echo w3m-spinner-map-help-echo)))
+			(favicon
+			 (propertize
+			  " "
+			  'display favicon
+			  'face face
+			  'local-map keymap
+			  'help-echo title))))
+		breadth (cond (icon width)
+			      (graphic (+ 2 width))
+			      (t width)))
+	  (push
+	   (list
+	    icon
 	    (propertize
-	     (concat (if favicon
-			 (propertize "  " 'display favicon)
-		       "  ")
-		     (if (and (> width 0)
-			      (> (string-width title) width))
-			 (if (> width 6)
-			     (concat
-			      (truncate-string-to-width
-			       title
-			       (max 0 (- width 3)))
-			      "...")
-			   (truncate-string-to-width title width))
-		       (concat title (make-string (max 0
-						       (- 
-							width
-							(string-width title)))
-						  ?\ )))
-		     "  ")
-	     'mouse-face 'highlight
-	     'face (if (progn (set-buffer buffer) w3m-current-process)
-		       (if (eq buffer current)
-			   'w3m-tab-selected-retrieving-face
-			 'w3m-tab-unselected-retrieving-face)
-		     (if (eq buffer current)
-			 'w3m-tab-selected-face
-		       'w3m-tab-unselected-face))
-	     'local-map (w3m-tab-make-keymap buffer)
-	     'help-echo title)))
-	buffers
-	(propertize " " 'face 'w3m-tab-background-face)))
-     (propertize (make-string (window-width) ?\ )
-		 'face 'w3m-tab-background-face))))
+	     (concat
+	      (when graphic w3m-tab-half-space)
+	      (if (and (> (string-width title) breadth)
+		       (> breadth 6))
+		  (truncate-string-to-width
+		   (concat (truncate-string-to-width title (- breadth 3))
+			   "...")
+		   breadth nil ?.)
+		(truncate-string-to-width title breadth nil ?\ )))
+	     'face face
+	     'local-map keymap
+	     'help-echo title)
+	    w3m-tab-separator)
+	   line))
+	(setq w3m-tab-line-format
+	      (concat (apply 'concat (apply 'nconc line))
+		      (propertize (make-string (window-width) ?\ )
+				  'face 'w3m-tab-background-face))))))
 
 (defun w3m-update-tab-line ()
   "Update tab line."
   (when w3m-use-tab
     (set-cursor-color (frame-parameter (selected-frame) 'cursor-color))))
 
+(add-hook 'w3m-mode-setup-functions 'w3m-tab-make-keymap)
 (add-hook 'w3m-mode-setup-functions 'w3m-setup-header-line)
 (add-hook 'w3m-mode-setup-functions 'w3m-setup-widget-faces)
 
+;; Graphic icons.
+(defcustom w3m-space-before-modeline-icon ""
+  "String of space character(s) to be put in front of the mode-line icon.
+It may be better to use one or more spaces if you are using oblique or
+italic font in the modeline."
+  :group 'w3m
+  :type 'string)
+
+(defvar w3m-spinner-image-file nil
+  "Image file used to show a spinner in the header-line.")
+
+(defvar w3m-spinner-image-frames 3
+  "Number of frames which the spinner image contains.")
+
+(defvar w3m-spinner-image-index 0
+  "Counter used to rotate spinner images.")
+
+;; Images to be displayed in the modeline.
+(defvar w3m-modeline-process-status-on-icon nil)
+(defvar w3m-modeline-image-status-on-icon nil)
+(defvar w3m-modeline-status-off-icon nil)
+(defvar w3m-modeline-ssl-image-status-on-icon nil)
+(defvar w3m-modeline-ssl-status-off-icon nil)
+
+(defun w3m-initialize-graphic-icons (&optional force)
+  "Make icon images which will be displayed in the mode-line."
+  (interactive "P")
+  (let ((defs '((w3m-modeline-status-off-icon
+		 "state-00.xpm"
+		 w3m-modeline-status-off)
+		(w3m-modeline-image-status-on-icon
+		 "state-01.xpm"
+		 w3m-modeline-image-status-on)
+		(w3m-modeline-ssl-status-off-icon
+		 "state-10.xpm"
+		 w3m-modeline-ssl-status-off)
+		(w3m-modeline-ssl-image-status-on-icon
+		 "state-11.xpm"
+		 w3m-modeline-ssl-image-status-on)))
+	def icon file status keymap)
+    (while defs
+      (setq def (car defs)
+	    defs (cdr defs)
+	    icon (car def)
+	    file (expand-file-name (nth 1 def) w3m-icon-directory)
+	    status (nth 2 def))
+      (if (and window-system
+	       w3m-show-graphic-icons-in-mode-line
+	       (display-images-p)
+	       (image-type-available-p 'xpm)
+	       (file-exists-p file))
+	  (progn
+	    (when (or force (not (symbol-value icon)))
+	      (unless keymap
+		(setq keymap (make-mode-line-mouse-map 'mouse-2
+						       'w3m-reload-this-page)))
+	      (set icon (propertize
+			 "  "
+			 'display (create-image file 'xpm nil :ascent 'center)
+			 'local-map keymap
+			 'help-echo "mouse-2 reloads this page"))
+	      (put icon 'risky-local-variable t)
+	      (put status 'risky-local-variable t))
+	    (when (stringp (symbol-value status))
+	      ;; Save the original status strings as properties.
+	      (put status 'string (symbol-value status)))
+	    (set status (list "" 'w3m-space-before-modeline-icon icon)))
+	;; Don't use graphic icons.
+	(when (get status 'string)
+	  (set status (get status 'string)))))
+    ;; Spinner
+    (when (and (or force (not w3m-spinner-image-file))
+	       (image-type-available-p 'gif)
+	       (file-exists-p
+		(setq file (expand-file-name "spinner.gif"
+					     w3m-icon-directory))))
+      (setq w3m-spinner-image-file file)
+      (define-key (setq w3m-modeline-spinner-map (make-sparse-keymap))
+	[mode-line mouse-2]
+	'w3m-process-stop)
+      (put 'w3m-modeline-process-status-on 'risky-local-variable t)
+      (put 'w3m-modeline-process-status-on-icon 'risky-local-variable t))
+    (if (and window-system
+	     w3m-show-graphic-icons-in-mode-line
+	     w3m-spinner-image-file)
+	(progn
+	  (when (stringp w3m-modeline-process-status-on)
+	    ;; Save the original status strings as properties.
+	    (put 'w3m-modeline-process-status-on 'string
+		 w3m-modeline-process-status-on))
+	  (setq w3m-modeline-process-status-on
+		'(""
+		  w3m-space-before-modeline-icon
+		  w3m-modeline-process-status-on-icon)))
+      (when (get 'w3m-modeline-process-status-on 'string)
+	(setq w3m-modeline-process-status-on
+	      (get 'w3m-modeline-process-status-on 'string))))))
+
+(defun w3m-make-spinner-image ()
+  "Make an image used to show a spinner.
+It should be called periodically in order to spin the spinner."
+  (when w3m-spinner-image-file
+    (unless (< (incf w3m-spinner-image-index) w3m-spinner-image-frames)
+      (setq w3m-spinner-image-index 0))
+    (let ((image (create-image w3m-spinner-image-file 'gif nil
+			       :ascent 'center :mask 'heuristic
+			       :index w3m-spinner-image-index)))
+      (setq w3m-modeline-process-status-on-icon
+	    (propertize "  "
+			'display image
+			'local-map w3m-modeline-spinner-map
+			'help-echo w3m-spinner-map-help-echo))
+      image)))
+
 (provide 'w3m-e21)
 
 ;;; w3m-e21.el ends here
