@@ -1,6 +1,6 @@
 ;;; w3m-proc.el --- Functions and macros to control sub-processes
 
-;; Copyright (C) 2001 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
+;; Copyright (C) 2001, 2002, 2003 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 
 ;; Authors: TSUCHIYA Masatoshi <tsuchiya@namazu.org>,
 ;;          Shun-ichi GOTO     <gotoh@taiyo.co.jp>,
@@ -44,8 +44,7 @@
 (eval-and-compile
   (cond ((boundp 'MULE)
 	 (autoload 'read-passwd "w3m-om"))
-	((and (boundp 'emacs-major-version)
-	      (= emacs-major-version 19))
+	((= emacs-major-version 19)
 	 (autoload 'read-passwd "w3m-19"))))
 
 (eval-when-compile
@@ -60,7 +59,9 @@
   (defvar w3m-command-arguments)
   (defvar w3m-command-environment)
   (defvar w3m-async-exec)
-  (defvar w3m-process-connection-type))
+  (defvar w3m-process-connection-type)
+  (defvar w3m-process-modeline-format)
+  (defvar w3m-work-buffer-list))
 
 (defvar w3m-process-inhibit-quit t
   "`w3m-process-sentinel' binds `inhibit-quit' according to this variable.")
@@ -85,6 +86,10 @@
 (make-variable-buffer-local 'w3m-process-realm)
 (make-variable-buffer-local 'w3m-process-object)
 
+(defvar w3m-process-modeline-string nil
+  "Modeline string to show status of retrieving process.")
+(make-variable-buffer-local 'w3m-process-modeline-string)
+
 (defvar w3m-process-proxy-user nil "User name of the proxy server.")
 (defvar w3m-process-proxy-passwd nil "Password of the proxy server.")
 
@@ -106,13 +111,22 @@
 	 (temporary-file-directory
 	  (if (file-directory-p w3m-profile-directory)
 	      (file-name-as-directory w3m-profile-directory)
-	    temporary-file-directory))
+	    ,(if (featurep 'xemacs)
+		 ;; Though `temporary-file-directory' exists even in XEmacs,
+		 ;; that's only an imitation provided by APEL.
+		 '(temp-directory)
+	       'temporary-file-directory)))
 	 (default-directory
 	   (cond ((file-directory-p w3m-profile-directory)
 		  (file-name-as-directory w3m-profile-directory))
 		 ((file-directory-p (expand-file-name "~/"))
 		  (expand-file-name "~/"))
 		 (t temporary-file-directory))))
+     ;; XEmacs obtains tmp-dir from the `temp-directory' function of which
+     ;; return value can only be modified by the following env vars.
+     ,@(if (featurep 'xemacs)
+	   '((setenv "TEMP" temporary-file-directory) ;; Windoze
+	     (setenv "TMPDIR" temporary-file-directory))) ;; Un|x
      (dolist (pair ,alist)
        (setenv (car pair) (cdr pair)))
      ,@body))
@@ -146,22 +160,24 @@
   `(aref (cdr ,object) 3))
 
 (put 'w3m-process-handler-new 'edebug-form-spec '(form form form))
-(defmacro w3m-process-handler-new (buffer parent-buffer function)
-  `(vector ,buffer ,parent-buffer ,function))
+(defmacro w3m-process-handler-new (buffer parent-buffer functions)
+  `(vector ,buffer ,parent-buffer ,functions nil))
 (defmacro w3m-process-handler-buffer (handler)
   `(aref ,handler 0))
 (defmacro w3m-process-handler-parent-buffer (handler)
   `(aref ,handler 1))
-(defmacro w3m-process-handler-function (handler)
+(defmacro w3m-process-handler-functions (handler)
   `(aref ,handler 2))
+(defmacro w3m-process-handler-result (handler)
+  `(aref ,handler 3))
 
-(defun w3m-process-push (handler arguments)
+(defun w3m-process-push (handler command arguments)
   "Generate a new `w3m-process' object which is provided by HANDLER,
 ARGUMENTS and this buffer, regist it to `w3m-process-queue', and
 return it."
-  (let ((x (assoc (cons w3m-command arguments) w3m-process-queue)))
+  (let ((x (assoc (cons command arguments) w3m-process-queue)))
     (unless x
-      (setq x (w3m-process-new w3m-command arguments (current-buffer)))
+      (setq x (w3m-process-new command arguments (current-buffer)))
       (push x w3m-process-queue))
     (push (w3m-process-handler-new (current-buffer) w3m-current-buffer handler)
 	  (w3m-process-handlers x))
@@ -173,14 +189,14 @@ return it."
   (when (processp process)
     (set-process-filter process 'ignore)
     (set-process-sentinel process 'ignore)
-    (when (eq (process-status process) 'run)
+    (when (memq (process-status process) '(run stop))
       (kill-process process)
       (when w3m-process-kill-surely
-	(while (eq (process-status process) 'run)
+	(while (memq (process-status process) '(run stop))
 	  (sit-for 0.1))))))
 
 (defun w3m-process-start-process (object &optional no-sentinel)
-  "Start a process spcified by the OBJECT, return always nil.
+  "Start a process specified by the OBJECT, return always nil.
 When NO-SENTINEL is not equal to nil, all status changes of the
 generated asynchronous process is ignored.  Otherwise,
 `w3m-process-sentinel' is given to the process as the sentinel."
@@ -194,7 +210,8 @@ generated asynchronous process is ignored.  Otherwise,
 		 (proc (apply 'start-process command
 			      (current-buffer) command
 			      (w3m-process-arguments object)))
-		 (authinfo (w3m-url-authinfo w3m-current-url)))
+		 (authinfo (when w3m-current-url
+			     (w3m-url-authinfo w3m-current-url))))
 	    (setq w3m-process-user (car authinfo)
 		  w3m-process-passwd (cdr authinfo)
 		  w3m-process-realm nil)
@@ -206,20 +223,32 @@ generated asynchronous process is ignored.  Otherwise,
 	    (process-kill-without-query proc))))))
   nil)	;; The return value of `w3m-process-start-process'.
 
+(defun w3m-process-kill-stray-processes ()
+  "Kill stray processes."
+  (dolist (obj w3m-process-queue)
+    (if (buffer-name (w3m-process-buffer obj))
+	(save-excursion
+	  (set-buffer (w3m-process-buffer obj))
+	  (dolist (x (w3m-process-handlers w3m-process-object))
+	    (unless (buffer-name (w3m-process-handler-parent-buffer x))
+	      (setq w3m-process-queue (delq obj w3m-process-queue))
+	      (when (w3m-process-process obj)
+		(w3m-process-kill-process (w3m-process-process obj))))))
+      (setq w3m-process-queue (delq obj w3m-process-queue))
+      (when (w3m-process-process obj)
+	(w3m-process-kill-process (w3m-process-process obj))))))
+
 (defun w3m-process-start-queued-processes ()
   "Start a process which is registerd in `w3m-process-queue' if the
 number of current working processes is less than `w3m-process-max'."
+  (w3m-process-kill-stray-processes)
   (let ((num 0))
     (catch 'last
       (dolist (obj (reverse w3m-process-queue))
-	(if (buffer-name (w3m-process-buffer obj))
-	    (if (> (incf num) w3m-process-max)
-		(throw 'last nil)
-	      (w3m-process-start-process obj))
-	  ;; Something wrong has occuered ?
-	  (setq w3m-process-queue (delq obj w3m-process-queue))
-	  (when (w3m-process-process obj)
-	    (w3m-process-kill-process (w3m-process-process obj))))))))
+	(when (buffer-name (w3m-process-buffer obj))
+	  (if (> (incf num) w3m-process-max)
+	      (throw 'last nil)
+	    (w3m-process-start-process obj)))))))
 
 (defun w3m-process-stop (buffer)
   "Remove handlers related to the buffer BUFFER, and stop processes
@@ -264,9 +293,22 @@ which have no handler."
 		     (dolist (handler (w3m-process-handlers obj))
 		       (w3m-kill-buffer (w3m-process-handler-buffer handler)))
 		     nil)))
-	       w3m-process-queue))
-	w3m-current-process nil)
-  (w3m-process-start-queued-processes))
+	       w3m-process-queue)))
+  (when (buffer-name buffer)
+    (with-current-buffer buffer
+      (setq w3m-current-process nil)))
+  (w3m-process-start-queued-processes)
+  (w3m-static-when (boundp 'header-line-format)
+    ;; Redisplay the header-line.
+    (run-at-time 0.5 nil
+		 (lambda (buffer)
+		   (if (and (buffer-live-p buffer)
+			    (eq (get-buffer-window buffer t)
+				(selected-window)))
+		       (let ((window-min-height 0))
+			 (shrink-window 1)
+			 (enlarge-window 1))))
+		 buffer)))
 
 (defun w3m-process-shutdown ()
   (let ((list w3m-process-queue))
@@ -283,9 +325,11 @@ which have no handler."
   "Generate the null handler, and evaluate BODY.
 When BODY is evaluated, the local variable `handler' keeps the null
 handler."
-  `(let ((handler (symbol-function 'identity)))
-     ,@body
-     (w3m-process-start-queued-processes)))
+  (let ((var (gensym "--tempvar--")))
+    `(let ((,var (let (handler) ,@body)))
+       (when (w3m-process-p ,var)
+	 (w3m-process-start-process ,var))
+       ,var)))
 (put 'w3m-process-with-null-handler 'lisp-indent-function 0)
 (put 'w3m-process-with-null-handler 'edebug-form-spec '(body))
 
@@ -293,37 +337,71 @@ handler."
 (put 'w3m-process-timeout 'error-conditions '(error w3m-process-timeout))
 (put 'w3m-process-timeout 'error-message "Time out")
 
+(defsubst w3m-process-error-handler (error-data process)
+  (setq w3m-process-queue (delq process w3m-process-queue))
+  (w3m-process-kill-process (w3m-process-process process))
+  (signal (car error-data) (cdr error-data)))
+
+(defun w3m-process-wait-process (process seconds)
+  "Wait for SECONDS seconds or until PROCESS will exit.
+Returns the exit status of the PROCESS when it exit normally,
+otherwise returns nil."
+  (catch 'timeout
+    (let ((start (current-time)))
+      (while (or (w3m-static-if (and (not (featurep 'xemacs))
+				     (>= emacs-major-version 21))
+		     (accept-process-output process 1)
+		   (not (sit-for 1)))
+		 (memq (process-status process) '(run stop)))
+	(and seconds
+	     (< seconds (w3m-time-lapse-seconds start (current-time)))
+	     (throw 'timeout nil)))
+      (w3m-static-if (and (not (featurep 'xemacs))
+			  (>= emacs-major-version 21))
+	  (accept-process-output process 0 0)
+	(sit-for 0))
+      (process-exit-status process))))
+
+(defun w3m-process-start-and-wait (w3m-current-process wait-function)
+  (while (w3m-process-p w3m-current-process)
+    (condition-case error
+	(let (w3m-process-inhibit-quit inhibit-quit)
+	  ;; No sentinel function is registered and the process
+	  ;; sentinel function is called from this macro, in order to
+	  ;; avoid the dead-locking which occurs when this macro is
+	  ;; called in the environment that `w3m-process-sentinel' is
+	  ;; evaluated.
+	  (w3m-process-start-process w3m-current-process t)
+	  (unless (w3m-process-wait-process
+		   (w3m-process-process w3m-current-process)
+		   w3m-process-timeout)
+	    (w3m-process-error-handler (cons 'w3m-process-timeout nil)
+				       w3m-current-process)))
+      (quit (w3m-process-error-handler error w3m-current-process)))
+    (w3m-process-sentinel (w3m-process-process w3m-current-process)
+			  "finished\n" t)
+    (setq w3m-current-process
+	  (catch 'result
+	    (dolist (handler (w3m-process-handlers w3m-current-process))
+	      (when (memq wait-function
+			  (w3m-process-handler-functions handler))
+		(throw 'result (w3m-process-handler-result handler))))
+	    (w3m-process-error-handler (cons 'error "Can't find wait handler")
+				       w3m-current-process))))
+  w3m-current-process)
+
 (defmacro w3m-process-with-wait-handler (&rest body)
   "Generate the waiting handler, and evaluate BODY.
 When BODY is evaluated, the local variable `handler' keeps the handler
 which will wait for the end of the evaluation."
   (let ((result (gensym "--result--"))
-	(start (gensym "--start--")))
-    `(let (,result)
-       (when (w3m-process-p
-	      (setq ,result
-		    (let ((handler (lambda (x) (setq ,result x))))
-		      ,@body)))
-	 (let ((,start (current-time))
-	       (w3m-current-process ,result)
-	       w3m-process-inhibit-quit inhibit-quit)
-	   ;; No sentinel function is registered and the process
-	   ;; sentinel function is called from this macro, in order to
-	   ;; avoid the dead-locking which occurs when this macro is
-	   ;; called in the environment that `w3m-process-sentinel' is
-	   ;; evaluated.
-	   (w3m-process-start-process ,result t)
-	   (while (eq (process-status (w3m-process-process ,result)) 'run)
-	     (accept-process-output nil 0 200)
-	     (and w3m-process-timeout
-		  (< w3m-process-timeout
-		     (w3m-time-lapse-seconds ,start (current-time)))
-		  (progn
-		    (setq w3m-process-queue (delq ,result w3m-process-queue))
-		    (w3m-process-kill-process (w3m-process-process ,result))
-		    (signal 'w3m-process-timeout nil)))))
-	 (w3m-process-sentinel (w3m-process-process ,result) "finished\n"))
-       ,result)))
+	(wait-function (gensym "--wait-function--")))
+    `(let ((,result)
+	   (,wait-function (make-symbol "wait-function")))
+       (fset ,wait-function 'identity)
+       (w3m-process-start-and-wait (let ((handler (list ,wait-function)))
+				     ,@body)
+				   ,wait-function))))
 (put 'w3m-process-with-wait-handler 'lisp-indent-function 0)
 (put 'w3m-process-with-wait-handler 'edebug-form-spec '(body))
 
@@ -370,32 +448,19 @@ evaluated at the same time, and this macro returns the result of the
 body BODY."
   (let ((var (or (car spec) (gensym "--tempvar--")))
 	(form (cdr spec))
-	(this-handler (gensym "--this-handler--")))
-    `(let ((,this-handler handler))
-       (labels ((post-body (,var handler) ,@body)
-		(post-handler
-		 (,var handler)
-		 (if (w3m-process-p (setq ,var (post-body ,var handler)))
-		     ;; The generated async process will be started at
-		     ;; the end of `w3m-process-sentinel', so that
-		     ;; there is nothing to do at this part.
-		     nil
-		   (funcall (or handler (function identity)) ,var))))
-	 (let ((,var
-		(let ((handler
-		       (list 'lambda (list ',var)
-			     (list 'post-handler ',var ,this-handler))))
-		  ,@form)))
-	   (if (w3m-process-p ,var)
-	       (if ,this-handler
+	(post-function (gensym "--post-function--")))
+    `(let ((,post-function (lambda (,var) ,@body)))
+       (let ((,var (let ((handler (cons ,post-function handler)))
+		     ,@form)))
+	 (if (w3m-process-p ,var)
+	     (if handler
+		 ,var
+	       (w3m-process-start-process ,var))
+	   (if (w3m-process-p (setq ,var (funcall ,post-function ,var)))
+	       (if handler
 		   ,var
 		 (w3m-process-start-process ,var))
-	     (if (w3m-process-p
-		  (setq ,var (post-body ,var ,this-handler)))
-		 (if ,this-handler
-		     ,var
-		   (w3m-process-start-process ,var))
-	       ,var)))))))
+	     ,var))))))
 (put 'w3m-process-do 'lisp-indent-function 1)
 (put 'w3m-process-do 'edebug-form-spec '((symbolp form) def-body))
 
@@ -405,57 +470,57 @@ Like `w3m-process-do', but the form FORM and the body BODY are
 evaluated in a temporary buffer."
   (let ((var (or (car spec) (gensym "--tempvar--")))
 	(form (cdr spec))
-	(this-handler (gensym "--this-handler--"))
-	(temp-buffer (gensym "--temp-buffer--")))
-    `(let ((,this-handler handler)
-	   (,temp-buffer
-	    (w3m-get-buffer-create
-	     (generate-new-buffer-name w3m-work-buffer-name))))
-       (labels ((post-body (,var handler ,temp-buffer)
-			   (unwind-protect
-			       (with-current-buffer ,temp-buffer
-				 ,@body)
-			     (w3m-kill-buffer ,temp-buffer)))
-		(post-handler (,var handler ,temp-buffer)
-			      (unless (w3m-process-p
-				       (setq ,var (post-body ,var handler
-							     ,temp-buffer)))
-				(funcall (or handler (function identity))
-					 ,var))))
-	 (let ((,var
-		(let ((handler
-		       (list 'lambda (list ',var)
-			     (list 'post-handler ',var
-				   ,this-handler ,temp-buffer))))
-		  (with-current-buffer ,temp-buffer ,@form))))
+	(post-body (gensym "--post-body--"))
+	(post-handler (gensym "--post-handler--"))
+	(temp-buffer (gensym "--temp-buffer--"))
+	(current-buffer (gensym "--current-buffer--")))
+    `(lexical-let ((,temp-buffer
+		    (w3m-get-buffer-create
+		     (generate-new-buffer-name w3m-work-buffer-name)))
+		   (,current-buffer (current-buffer)))
+       (labels ((,post-body (,var)
+			    (when (buffer-name ,temp-buffer)
+			      (set-buffer ,temp-buffer))
+			    ,@body)
+		(,post-handler (,var)
+			       (w3m-kill-buffer ,temp-buffer)
+			       (when (buffer-name ,current-buffer)
+				 (set-buffer ,current-buffer))
+			       ,var))
+	 (let ((,var (let ((handler
+			    (cons ',post-body (cons ',post-handler handler))))
+		       (with-current-buffer ,temp-buffer ,@form))))
 	   (if (w3m-process-p ,var)
-	       (if ,this-handler
+	       (if handler
 		   ,var
 		 (w3m-process-start-process ,var))
 	     (if (w3m-process-p
-		  (setq ,var (post-body ,var ,this-handler ,temp-buffer)))
-		 (if ,this-handler
+		  (setq ,var (let ((handler (cons ',post-handler handler)))
+			       (,post-body ,var))))
+		 (if handler
 		     ,var
 		   (w3m-process-start-process ,var))
-	       ,var)))))))
+	       (,post-handler ,var))))))))
 (put 'w3m-process-do-with-temp-buffer 'lisp-indent-function 1)
 (put 'w3m-process-do-with-temp-buffer 'edebug-form-spec
      '((symbolp form) def-body))
 
 
-(defun w3m-process-start (handler &rest arguments)
-  "Run `w3m-command' with HANDLER and ARGUMENTS."
-  (setq arguments (append w3m-command-arguments arguments))
+(defun w3m-process-start (handler command arguments)
+  "Run COMMAND with ARGUMENTS, and eval HANDLER asynchronously."
   (if w3m-async-exec
       (w3m-process-do
-	  (exit-status (w3m-process-push handler arguments))
+	  (exit-status (w3m-process-push handler command arguments))
 	(w3m-process-start-after exit-status))
     (w3m-process-start-after
      (w3m-process-with-coding-system
        (w3m-process-with-environment w3m-command-environment
-	 (apply 'call-process w3m-command nil t nil arguments))))))
+	 (apply 'call-process command nil t nil arguments))))))
 
 (defun w3m-process-start-after (exit-status)
+  (when w3m-current-buffer
+    (with-current-buffer w3m-current-buffer
+      (setq w3m-process-modeline-string nil)))
   (cond
    ((numberp exit-status)
     (zerop (setq w3m-process-exit-status exit-status)))
@@ -466,14 +531,13 @@ evaluated in a temporary buffer."
 	  (string-as-multibyte (format "%s" exit-status)))
     nil)))
 
-(defun w3m-process-sentinel (process event)
+(defun w3m-process-sentinel (process event &optional ignore-queue)
   ;; Ensure that this function will be never called repeatedly.
   (set-process-sentinel process 'ignore)
   (let ((inhibit-quit w3m-process-inhibit-quit))
     (unwind-protect
 	(if (buffer-name (process-buffer process))
-	    (save-current-buffer
-	      (set-buffer (process-buffer process))
+	    (with-current-buffer (process-buffer process)
 	      (setq w3m-process-queue
 		    (delq w3m-process-object w3m-process-queue))
 	      (let ((exit-status (process-exit-status process))
@@ -493,12 +557,19 @@ evaluated in a temporary buffer."
 		    (set-buffer (w3m-process-handler-buffer x))
 		    (let ((w3m-process-exit-status)
 			  (w3m-current-buffer
-			   (w3m-process-handler-parent-buffer x)))
+			   (w3m-process-handler-parent-buffer x))
+			  (handler
+			   (w3m-process-handler-functions x))
+			  (exit-status exit-status))
 		      (when realm
 			(w3m-process-set-authinfo w3m-current-url
 						  realm user passwd))
-		      (funcall (w3m-process-handler-function x)
-			       exit-status))))))
+		      (while (and handler
+				  (not (w3m-process-p
+					(setq exit-status
+					      (funcall (pop handler)
+						       exit-status))))))
+		      (setf (w3m-process-handler-result x) exit-status))))))
 	  ;; Something wrong has been occured.
 	  (catch 'last
 	    (dolist (obj w3m-process-queue)
@@ -506,7 +577,8 @@ evaluated in a temporary buffer."
 		(setq w3m-process-queue (delq obj w3m-process-queue))
 		(throw 'last nil)))))
       (delete-process process)
-      (w3m-process-start-queued-processes))))
+      (unless ignore-queue
+	(w3m-process-start-queued-processes)))))
 
 (defun w3m-process-filter (process string)
   (when (buffer-name (process-buffer process))
@@ -576,16 +648,34 @@ evaluated in a temporary buffer."
 					   (match-beginning 1))))
 	    (ignore-errors
 	      (process-send-string process
-				   (concat w3m-process-user "\n"))))))))))
+				   (concat w3m-process-user "\n"))))
+	   ((progn
+	      (or (search-forward "\nW3m-current-url:" nil t)
+		  (goto-char (process-mark process)))
+	      (re-search-backward
+	       "^W3m-\\(in-\\)?progress: \\([.0-9]+/[.0-9]+[a-zA-Z]?b\\)$"
+	       nil t))
+	    (let ((str (w3m-process-modeline-format (match-string 2)))
+		  (buf))
+	      (save-current-buffer
+		(dolist (handler (w3m-process-handlers w3m-process-object))
+		  (when (setq buf (w3m-process-handler-parent-buffer handler))
+		    (if (buffer-name buf)
+			(progn
+			  (set-buffer buf)
+			  (setq w3m-process-modeline-string str))
+		      (w3m-process-kill-stray-processes)))))))))))))
 
-(defun w3m-process-get-server-root (url)
-  "Extract a server root from URL."
-  (when (string-match "\\`about://[^/?#]+/" url)
-    (setq url (substring url (match-end 0))))
-  (setq url (w3m-url-strip-authinfo url))
-  (if (string-match "\\`[^:/?#]+://\\([^/?#]+\\)" url)
-      (downcase (match-string 1 url))
-    url))
+(defun w3m-process-modeline-format (str)
+  (ignore-errors
+    (cond
+     ((stringp w3m-process-modeline-format)
+      (format w3m-process-modeline-format
+	      (if (string-match "/0\\([a-zA-Z]?b\\)\\'" str)
+		  (replace-match "\\1" t nil str)
+		str)))
+     ((functionp w3m-process-modeline-format)
+      (funcall w3m-process-modeline-format str)))))
 
 ;; w3m-process-authinfo-alist has an association list as below format.
 ;; (("root1" ("realm11" ("user11" . "pass11")
@@ -593,7 +683,7 @@ evaluated in a temporary buffer."
 ;;           ("realm12" ("user13" . "pass13")))
 ;;  ("root2" ("realm21" ("user21" . "pass21"))))
 (defun w3m-process-set-authinfo (url realm username password)
-  (let (x y z (root (w3m-process-get-server-root url)))
+  (let (x y z (root (w3m-get-server-hostname url)))
     (if (setq x (assoc root w3m-process-authinfo-alist))
 	(if (setq y (assoc realm x))
 	    (if (setq z (assoc username y))
@@ -610,7 +700,7 @@ evaluated in a temporary buffer."
 
 (defun w3m-process-read-user (url &optional realm ignore-history)
   "Read a user name for URL and REALM."
-  (let* ((root (when (stringp url) (w3m-process-get-server-root url)))
+  (let* ((root (when (stringp url) (w3m-get-server-hostname url)))
 	 (ident (or realm root))
 	 (alist))
     (if (and (not ignore-history)
@@ -631,7 +721,7 @@ evaluated in a temporary buffer."
 
 (defun w3m-process-read-passwd (url &optional realm username ignore-history)
   "Read a password for URL, REALM, and USERNAME."
-  (let* ((root (when (stringp url) (w3m-process-get-server-root url)))
+  (let* ((root (when (stringp url) (w3m-get-server-hostname url)))
 	 (ident (or realm root))
 	 (pass (cdr (assoc username
 			   (cdr (assoc realm
@@ -656,7 +746,7 @@ evaluated in a temporary buffer."
 NOTE: This function is designed to avoid annoying questions.  So when
 the same questions is reasked, its previous answer is reused without
 prompt."
-  (let (elem answer (root (w3m-process-get-server-root url)))
+  (let (elem answer (root (w3m-get-server-hostname url)))
     (if (setq elem (assoc root w3m-process-accept-alist))
 	(if (member prompt (cdr elem))
 	    ;; When the same question has been asked, the previous
