@@ -1,6 +1,6 @@
 ;;; w3m-util.el --- Utility macros and functions for emacs-w3m
 
-;; Copyright (C) 2001, 2002 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
+;; Copyright (C) 2001, 2002, 2003 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 
 ;; Authors: TSUCHIYA Masatoshi <tsuchiya@namazu.org>,
 ;;          Shun-ichi GOTO     <gotoh@taiyo.co.jp>,
@@ -41,13 +41,20 @@
   (require 'cl)
   ;; Variables and functions which are used in the following inline
   ;; functions.  They should be defined in the other module at run-time.
-  (defvar w3m-current-url)
-  (defvar w3m-current-title)
-  (defvar w3m-html-string-regexp)
-  (defvar w3m-work-buffer-list)
+  (defvar w3m-current-process)
   (defvar w3m-current-refresh)
+  (defvar w3m-current-title)
+  (defvar w3m-current-url)
+  (defvar w3m-pop-up-frames)
+  (defvar w3m-pop-up-windows)
+  (defvar w3m-popup-frame-parameters)
   (defvar w3m-refresh-timer)
-  (defvar w3m-use-refresh))
+  (defvar w3m-select-buffer-name)
+  (defvar w3m-use-refresh)
+  (defvar w3m-use-tab)
+  (defvar w3m-work-buffer-list)
+  (unless (fboundp 'select-frame-set-input-focus)
+    (defalias 'select-frame-set-input-focus 'ignore)))
 
 (eval-and-compile
   (dont-compile
@@ -63,17 +70,16 @@
 (eval-and-compile
   (cond
    ((or (featurep 'xemacs)
-	(not (boundp 'emacs-major-version))
 	(< emacs-major-version 20))
     (require 'poe)
     (require 'poem))))
 
 (eval-and-compile
-  (cond ((and (boundp 'emacs-major-version)
-	      (= emacs-major-version 19))
+  (cond ((= emacs-major-version 19)
 	 (autoload 'cancel-timer "timer")
 	 (require 'custom))))
 
+
 ;;; Things should be defined in advance:
 
 ;; (There are no objects so far.)
@@ -127,43 +133,82 @@ or `debug-on-quit' is non-nil."
 	 (if (featurep 'xemacs)
 	     ;; Default to start-closed and end-open in XEmacsen.
 	     '(list 'start-open t)
-	   ;; Default to front-nonsticky and rear-sticky in FSF Emacsen.
+	   ;; Default to front-nonsticky and rear-sticky in Emacsen.
 	   '(list 'rear-nonsticky t))))
     (` (add-text-properties (, start) (, end)
 			    (append (, non-stickies) (, props))
 			    (, object)))))
 
-(defmacro w3m-get-text-property-around (prop &optional position)
-  "Search for the text property PROP in the POSITION and return a value
-or nil.  If POSITION is omitted, searching is performed in the current
-cursor position and around there."
-  (if position
-      (` (get-text-property (, position) (, prop)))
-    (` (let ((position (point)))
-	 (or (get-text-property position (, prop))
-	     (and (not (bolp))
-		  (get-text-property (1- position) (, prop)))
-	     (and (not (eolp))
-		  (get-text-property (1+ position) (, prop))))))))
+(defmacro w3m-get-text-property-around (prop)
+  "Search for the text property PROP in one character before and behind
+the current position.  Return the value corresponding to PROP or nil.
+If PROP is not found at the current position, point will move to the
+position where PROP exists."
+  (` (let ((position (point))
+	   value)
+       (or (get-text-property position (, prop))
+	   (and (not (bolp))
+		(setq value (get-text-property (1- position) (, prop)))
+		(goto-char (1- position))
+		value)
+	   (and (not (eolp))
+		(setq value (get-text-property (1+ position) (, prop)))
+		(goto-char (1+ position))
+		value)))))
 
 (defmacro w3m-action (&optional position)
-  (` (w3m-get-text-property-around 'w3m-action (, position))))
+  "Return the value of the `w3m-action' property at the given POSITION.
+NOTE: If POSITION is omitted, it searches for the property in one
+character before and behind the current position, and point will move
+to the position where the property exists."
+  (if position
+      (` (get-text-property (, position) 'w3m-action))
+    (` (w3m-get-text-property-around 'w3m-action))))
+
 (defmacro w3m-anchor (&optional position)
-  (` (w3m-get-text-property-around 'w3m-href-anchor (, position))))
+  "Return the value of the `w3m-href-anchor' property at the given POSITION.
+NOTE: If POSITION is omitted, it searches for the property in one
+character before and behind the current position, and point will move
+to the position where the property exists."
+  (if position
+      (` (get-text-property (, position) 'w3m-href-anchor))
+    (` (w3m-get-text-property-around 'w3m-href-anchor))))
+
 (defmacro w3m-image (&optional position)
-  (` (w3m-get-text-property-around 'w3m-image (, position))))
-(defmacro w3m-image-scale (&optional position)
-  (` (w3m-get-text-property-around 'w3m-image-scale (, position))))
+  "Return the value of the `w3m-image' property at the given POSITION.
+NOTE: If POSITION is omitted, it searches for the property in one
+character before and behind the current position, and point will move
+to the position where the property exists."
+  (if position
+      (` (get-text-property (, position) 'w3m-image))
+    (` (w3m-get-text-property-around 'w3m-image))))
+
 (defmacro w3m-submit (&optional position)
-  (` (w3m-get-text-property-around 'w3m-submit (, position))))
+  "Return the value of the `w3m-submit' property at the given POSITION.
+NOTE: If POSITION is omitted, it searches for the property in one
+character before and behind the current position, and point will move
+to the position where the property exists."
+  (if position
+      (` (get-text-property (, position) 'w3m-submit))
+    (` (w3m-get-text-property-around 'w3m-submit))))
 
 (defmacro w3m-anchor-sequence (&optional position)
+  "Return the value of the `w3m-anchor-sequence' property at POSITION.
+If POSITION is omitted, the current position is assumed."
   (if position
       (` (get-text-property (, position) 'w3m-anchor-sequence))
-    (` (get-text-property (point) 'w3m-anchor-sequence))))
+    '(get-text-property (point) 'w3m-anchor-sequence)))
+
 
 ;;; Attributes:
 
+(eval-and-compile
+  ;; `eval-and-compile' is necessary since the value of the constant
+  ;; is referred to at the compile time.
+  (defconst w3m-html-string-regexp
+    "\\(\"\\([^\"]+\\)\"\\|'\\([^\']+\\)'\\|[^\"\'<> \t\r\f\n]*\\)"
+    "Regexp matching a string of the field-value like <a href=\"VALUE\">."))
+
 (put 'w3m-parse-attributes 'lisp-indent-function '1)
 (def-edebug-spec w3m-parse-attributes
   ((&rest &or (symbolp &optional symbolp) symbolp) body))
@@ -251,19 +296,23 @@ An argument of nil means kill the current buffer."
       (kill-buffer buf)))
   (setq w3m-work-buffer-list nil))
 
+(defsubst w3m-current-title ()
+  "Return the title of the current buffer."
+  (cond
+   ((and (stringp w3m-current-title)
+	 (not (string= w3m-current-title "<no-title>")))
+    w3m-current-title)
+   ((stringp w3m-current-url)
+    (directory-file-name
+     (if (string-match "^[^/:]+:/+" w3m-current-url)
+	 (substring w3m-current-url (match-end 0))
+       w3m-current-url)))
+   (t "<no-title>")))
+
 (defsubst w3m-buffer-title (buffer)
   "Return the title of the buffer BUFFER."
   (with-current-buffer buffer
-    (cond
-     ((and (stringp w3m-current-title)
-	   (not (string= w3m-current-title "<no-title>")))
-      w3m-current-title)
-     ((stringp w3m-current-url)
-      (directory-file-name
-       (if (string-match "^[^/:]+:/+" w3m-current-url)
-	   (substring w3m-current-url (match-end 0))
-	 w3m-current-url)))
-     (t "<no-title>"))))
+    (w3m-current-title)))
 
 (defsubst w3m-buffer-number (buffer)
   (when (and (bufferp buffer)
@@ -301,22 +350,244 @@ An argument of nil means kill the current buffer."
 	(string< (car x) y))
     (string< x y)))
 
-(defsubst w3m-list-buffers (&optional nosort)
-  "Return list of w3m-mode buffers."
-  (if nosort
-      (save-current-buffer
-	(delq nil
-	      (mapcar
-	       (lambda (buffer)
-		 (set-buffer buffer)
-		 (when (eq major-mode 'w3m-mode) buffer))
-	       (buffer-list))))
-    (sort (w3m-list-buffers t)
-	  (function w3m-buffer-name-lessp))))
+(defun w3m-list-buffers (&optional nosort)
+  "Return a list of buffers in which emacs-w3m sessions are open.
+If the optional NOSORT is nil, the list is sorted in the order of
+buffer names."
+  (let ((buffers (buffer-list))
+	buffer rest)
+    (save-current-buffer
+      (while buffers
+	(set-buffer (setq buffer (pop buffers)))
+	(when (eq major-mode 'w3m-mode)
+	  (push buffer rest))))
+    (if nosort
+	(nreverse rest)
+      (sort rest #'w3m-buffer-name-lessp))))
+
+
+;;; Pop up and delete buffers, windows or frames:
+
+(defmacro w3m-popup-frame-parameters ()
+  "Return a pop-up frame plist if this file is compiled for XEmacs,
+otherwise return an alist."
+  (if (featurep 'xemacs)
+      '(let ((params (or w3m-popup-frame-parameters pop-up-frame-plist)))
+	 (if (consp (car-safe params))
+	     (alist-to-plist params)
+	   params))
+    '(let ((params (or w3m-popup-frame-parameters pop-up-frame-alist))
+	   alist)
+       (if (consp (car-safe params))
+	   params
+	 (while params
+	   (push (cons (car params) (cdr params)) alist)
+	   (setq params (cddr params)))
+	 (nreverse alist)))))
+
+(defmacro w3m-popup-frame-p ()
+  "Return non-nil if `w3m-pop-up-frames' is non-nil and Emacs really
+supports separate frames."
+  (if (featurep 'xemacs)
+      '(and w3m-pop-up-frames (device-on-window-system-p))
+    '(and w3m-pop-up-frames window-system)))
+
+(defmacro w3m-use-tab-p ()
+  "Return non-nil if `w3m-use-tab' is non-nil and Emacs really supports
+the tabs line."
+  (cond ((featurep 'xemacs)
+	 '(and w3m-use-tab (device-on-window-system-p)))
+	((<= emacs-major-version 19)
+	 nil)
+	(t
+	 '(and w3m-use-tab (>= emacs-major-version 21)))))
+
+(defmacro w3m-popup-window-p ()
+  "Return non-nil if `w3m-pop-up-windows' is non-nil and the present
+situation allows it."
+  (cond ((featurep 'xemacs)
+	 '(and w3m-pop-up-windows
+	       (not (w3m-use-tab-p))
+	       (not (get-buffer-window w3m-select-buffer-name))))
+	((<= emacs-major-version 19)
+	 '(and w3m-pop-up-windows
+	       (not (get-buffer-window w3m-select-buffer-name))))
+	(t
+	 '(and w3m-pop-up-windows
+	       (or (< emacs-major-version 21)
+		   (not (w3m-use-tab-p)))
+	       (not (get-buffer-window w3m-select-buffer-name))))))
+
+(defvar w3m-initial-frames nil
+  "Variable used to keep a list of the frame-IDs when emacs-w3m sessions
+are popped-up as new frames.  This variable is used for the control
+for not deleting frames made for aims other than emacs-w3m sessions.")
+(make-variable-buffer-local 'w3m-initial-frames)
+
+(defun w3m-popup-buffer (buffer)
+  "Pop up BUFFER as a new window or a new frame
+according to `w3m-pop-up-windows' and `w3m-pop-up-frames' (which see)."
+  (let ((window (get-buffer-window buffer t))
+	(oframe (selected-frame))
+	(popup-frame-p (w3m-popup-frame-p))
+	frame pop-up-frames buffers other)
+    (if (setq
+	 pop-up-frames
+	 (if window ;; The window showing BUFFER already exists.
+	     ;; Don't pop up a new frame if it is just the current frame.
+	     (not (eq (setq frame (window-frame window)) oframe))
+	   ;; There is no window for BUFFER, so look for the existing
+	   ;; emacs-w3m window if the tabs line is enabled or the
+	   ;; selection window exists (i.e., we can reuse it).
+	   (if (or (w3m-use-tab-p)
+		   (get-buffer-window w3m-select-buffer-name t))
+	       (progn
+		 (setq buffers (delq buffer (w3m-list-buffers t)))
+		 (while (and (not window)
+			     buffers)
+		   (setq window
+			 (get-buffer-window (setq other (pop buffers)) t)))
+		 (if window ;; The window showing another buffer exists.
+		     (not (eq (setq frame (window-frame window)) oframe))
+		   (setq other nil)
+		   ;; There is no window after all, so leave to the value
+		   ;; of `w3m-pop-up-frames' whether to pop up a new frame.
+		   popup-frame-p))
+	     ;; Ditto.
+	     popup-frame-p)))
+	(progn
+	  (cond (other
+		 ;; Pop up another emacs-w3m buffer and switch to BUFFER.
+		 (pop-to-buffer other)
+		 ;; Change the value for BUFFER's `w3m-initial-frames'.
+		 (setq w3m-initial-frames
+		       (prog1
+			   (copy-sequence w3m-initial-frames)
+			 (switch-to-buffer buffer))))
+		(frame
+		 ;; Pop up the existing frame which shows BUFFER.
+		 (pop-to-buffer buffer))
+		(t
+		 ;; Pop up a new frame.
+		 (let* ((pop-up-frame-alist (w3m-popup-frame-parameters))
+			(pop-up-frame-plist pop-up-frame-alist))
+		   (pop-to-buffer buffer))
+		 (setq frame (window-frame (get-buffer-window buffer t)))))
+	  ;; Raise, select and focus the frame.
+	  (if (fboundp 'select-frame-set-input-focus)
+	      (select-frame-set-input-focus frame)
+	    (raise-frame frame)
+	    (select-frame frame)
+	    (focus-frame frame)))
+      ;; Simply switch to BUFFER in the current frame.
+      (if (w3m-popup-window-p)
+	  (let ((pop-up-windows t))
+	    (pop-to-buffer buffer))
+	(switch-to-buffer buffer)))))
+
+(eval-when-compile
+  (when (and (fboundp 'select-frame-set-input-focus)
+	     (eq (symbol-function 'select-frame-set-input-focus) 'ignore))
+    (fmakunbound 'select-frame-set-input-focus)))
+
+(defun w3m-add-w3m-initial-frames (&optional frame)
+  "Add a frame to `w3m-initial-frames' when it is newly created for the
+emacs-w3m session.  This function is added to the hook which is
+different with the version of Emacs as follows:
+
+XEmacs          create-frame-hook
+Emacs 20,21     after-make-frame-functions
+Emacs 19        after-make-frame-hook\
+"
+  (unless frame
+    (setq frame (selected-frame)))
+  (with-current-buffer (window-buffer (frame-first-window frame))
+    (when (eq major-mode 'w3m-mode)
+      (push frame w3m-initial-frames))))
+
+(add-hook (cond ((featurep 'xemacs)
+		 'create-frame-hook)
+		((>= emacs-major-version 20)
+		 'after-make-frame-functions)
+		((= emacs-major-version 19)
+		 'after-make-frame-hook))
+	  'w3m-add-w3m-initial-frames)
+
+(defun w3m-delete-frames-and-windows (&optional exception)
+  "Delete all frames and windows related to emacs-w3m buffers.
+If EXCEPTION is a buffer, a window or a frame, it and related visible
+objects will not be deleted.  There are special cases; the following
+objects will not be deleted:
+
+1. The sole frame in the display device.
+2. Frames created not for emacs-w3m sessions.
+3. Frames showing not only emacs-w3m sessions but also other windows.\
+"
+  (let ((buffers (delq exception (w3m-list-buffers t)))
+	buffer windows window frame one-window-p flag)
+    (save-current-buffer
+      (while buffers
+	(setq buffer (pop buffers)
+	      windows (delq exception
+			    (get-buffer-window-list buffer 'no-minibuf t)))
+	(set-buffer buffer)
+	(while windows
+	  (setq window (pop windows)
+		frame (window-frame window))
+	  (when (and frame
+		     (not (eq frame exception)))
+	    (setq one-window-p
+		  (w3m-static-if (featurep 'xemacs)
+		      (one-window-p t frame)
+		    ;; Emulate XEmacs version's `one-window-p'.
+		    (setq flag nil)
+		    (catch 'exceeded
+		      (walk-windows (lambda (w)
+				      (when (eq (window-frame w) frame)
+					(if flag
+					    (throw 'exceeded nil)
+					  (setq flag t))))
+				    'no-minibuf t)
+		      flag)))
+	    (if (and (memq frame w3m-initial-frames)
+		     (not (eq (next-frame frame) frame)))
+		(if (or
+		     ;; A frame having the sole window can be deleted.
+		     one-window-p
+		     ;; Also a frame having only windows for emacs-w3m
+		     ;; sessions or the buffer selection can be deleted.
+		     (progn
+		       (setq flag t)
+		       (walk-windows
+			(lambda (w)
+			  (when flag
+			    (if (eq w exception)
+				(setq flag nil)
+			      (set-buffer (window-buffer w))
+			      (setq flag (memq major-mode
+					       '(w3m-mode
+						 w3m-select-buffer-mode))))))
+			'no-minibuf)
+		       (set-buffer buffer)
+		       flag))
+		    (progn
+		      (setq w3m-initial-frames (delq frame
+						     w3m-initial-frames))
+		      (delete-frame frame))
+		  (delete-window window))
+	      (unless one-window-p
+		(delete-window window)))))))))
 
 
 ;;; Miscellaneous:
 
+(defconst w3m-url-fallback-base "http:///")
+(defconst w3m-url-invalid-regexp "\\`http:///")
+
+(defsubst w3m-url-valid (url)
+  (and url (not (string-match w3m-url-invalid-regexp url))
+       url))
+
 (defmacro w3m-tag-regexp-of (&rest names)
   "Return a regexp string, not a funtion form.  A regexp should match tags
 which are started with \"<\" and one of NAMES.  NAMES should be string
@@ -342,15 +613,31 @@ START and END are lists which represent time in Emacs-style."
      (cadr end)
      (- (cadr start))))
 
-(defsubst w3m-url-dtree-p (url)
-  "If URL points a 'w3m-dtree', return non-nil value.  Otherwise return
-nil."
-  (string-match "\\`about://dtree/" url))
+(defalias 'w3m-float-time
+  (if (fboundp 'float-time)
+      'float-time
+    (lambda (&optional specified-time)
+      "Return the current time, as a float number of seconds since the epoch.
+If an argument is given, it specifies a time to convert to float
+instead of the current time.  The argument should have the forms:
+ (HIGH . LOW) or (HIGH LOW USEC) or (HIGH LOW . USEC).
+
+WARNING: Since the result is floating point, it may not be exact.
+Do not use this function if precise time stamps are required."
+      (let ((time (or specified-time (current-time))))
+	(+ (* (car time) 65536.0)
+	   (cadr time)
+	   (cond ((consp (setq time (cddr time)))
+		  (/ (car time) 1000000.0))
+		 (time
+		  (/ time 1000000.0))
+		 (t
+		  0)))))))
 
 (defsubst w3m-url-local-p (url)
   "If URL points a file on the local system, return non-nil value.
 Otherwise return nil."
-  (string-match "^\\(file:\\|/\\|[a-zA-Z]:/\\)" url))
+  (string-match "\\`file:" url))
 
 (defconst w3m-url-authinfo-regexp
   "\\`\\([^:/?#]+:\\)?//\\([^/?#:]+\\)\\(:\\([^/?#@]+\\)\\)?@"
@@ -382,6 +669,15 @@ Otherwise return nil."
       (match-string 1 url)
     url))
 
+(defun w3m-get-server-hostname (url)
+  "Extract a server root from URL."
+  (when (string-match "\\`about://[^/?#]+/" url)
+    (setq url (substring url (match-end 0))))
+  (setq url (w3m-url-strip-authinfo url))
+  (if (string-match "\\`[^:/?#]+://\\([^/?#]+\\)" url)
+      (downcase (match-string 1 url))
+    url))
+
 (defsubst w3m-which-command (command)
   (when (stringp command)
     (if (and (file-name-absolute-p command)
@@ -451,26 +747,6 @@ Otherwise return nil."
 	(prin1 (apply 'concat (nreverse rest)) stream))
     (prin1 object stream)))
 
-(defvar w3m-display-message-enable-logging nil
-  "*If non-nil, preserve echo messages in the *Message* buffer.")
-
-(defmacro w3m-display-message (string &rest args)
-  "Like `message', except that message logging is controlled by the
-variable `w3m-display-message-enable-logging'."
-  (if (featurep 'xemacs)
-      (if args
-	  `(display-message (if w3m-display-message-enable-logging
-				'message
-			      'no-log)
-	     (format ,string ,@args))
-	`(display-message (if w3m-display-message-enable-logging
-			      'message
-			    'no-log)
-	   ,string))
-    `(let ((message-log-max (if w3m-display-message-enable-logging
-				message-log-max)))
-       (message ,string ,@args))))
-
 (defun w3m-display-progress-message (url)
   "Show \"Reading URL...\" message in the middle of a buffer."
   (insert (make-string (max 0 (/ (1- (window-height)) 2)) ?\n)
@@ -479,25 +755,9 @@ variable `w3m-display-message-enable-logging'."
   (let ((fill-column (window-width)))
     (center-region (point) (point-max)))
   (goto-char (point-min))
+  (put-text-property (point) (point-max) 'w3m-progress-message t)
   (sit-for 0))
 
-(if (featurep 'xemacs)
-    (defalias 'w3m-function-max-args 'function-max-args)
-  (eval-and-compile
-    (require 'advice))
-  (defun w3m-function-max-args (function)
-    "Return the maximum number of arguments a function may be called with.
-The function may be any form that can be passed to `funcall',
-any special form, or any macro.
-If the function takes an arbitrary number of arguments or is
-a built-in special form, nil is returned."
-    (let ((arglist (ad-arglist (if (symbolp function)
-				   (symbol-function function)
-				 function))))
-      (if (memq '&rest arglist)
-	  nil
-	(length (delq '&optional arglist))))))
-
 (defun w3m-modify-plist (plist &rest properties)
   "Change values in PLIST corresponding to PROPERTIES.  This is similar
 to `plist-put', but handles plural symbol and value pairs and remove
@@ -529,7 +789,7 @@ multibyteness of the buffer."
 		   (fg (face-foreground 'default)))
 	       (append (if bg `(:background ,bg))
 		       (if fg `(:foreground ,fg))))))
-  "The initial `default' face color spec.  Since `defface' under FSF Emacs
+  "The initial `default' face color spec.  Since `defface' under Emacs
 versions prior to 21 won't inherit the `dafault' face colors by default,
 we will use this value for the default `defface' color spec.")
 
@@ -543,6 +803,13 @@ we will use this value for the default `defface' color spec.")
 	  (setq value (cdr value))))
     (custom-initialize-set symbol value)))
 
+(defmacro w3m-activate-zmacs-regions ()
+  "Set t to `zmacs-region-stays' when a user invoke a command.
+This is the XEmacs specific macro."
+  (if (featurep 'xemacs)
+      '(if (interactive-p)
+	   (setq zmacs-region-stays t))))
+
 (provide 'w3m-util)
 
 ;;; w3m-util.el ends here
